import numpy as np
import warnings
from AnyQt.QtCore import Qt
from AnyQt.QtWidgets import QFormLayout
from functools import partial
from types import SimpleNamespace as namespace
from typing import Optional  # pylint: disable=unused-import

from Orange.data import Table, Domain
from Orange.data.util import array_equal
from Orange.misc import DistMatrix
from Orange.preprocess import preprocess
from Orange.projection import PCA
from Orange.projection import manifold
from Orange.widgets import gui
from Orange.widgets.settings import SettingProvider, ContextSetting
from Orange.widgets.utils.concurrents import TaskState, ConcurrentWidgetMixin
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.visualize.owscatterplotgraph import OWScatterPlotBase
from Orange.widgets.visualize.utils.widget import OWDataProjectionWidget
from Orange.widgets.widget import Msg
from orangewidget.utils.signals import Input

from Orange.i18n_config import *


def __(key):
    return i18n.t("widget.unsupervised.unsupervised.owtsne." + key)


_STEP_SIZE = 25
_MAX_PCA_COMPONENTS = 50
_DEFAULT_PCA_COMPONENTS = 20

INITIALIZATIONS = [("PCA", "pca"), ("Spectral", "spectral")]
DISTANCE_METRICS = [("Euclidean", "l2"), ("Manhattan", "l1"), ("Cosine", "cosine")]


class Task(namespace):
    """Completely determines the t-SNE task spec and intermediate results."""
    data = None  # type: Optional[Table]
    distance_matrix = None          # type: Optional[DistMatrix]

    preprocessed_data = None        # type: Optional[Table]normalize = None  # type: Optional[bool]
    normalized_data = None          # type: Optional[Table]

    use_pca_preprocessing = None    # type: Optional[bool]pca_components = None  # type: Optional[int]
    pca_projection = None  # type: Optional[Table]
    distance_metric = None          # type: Optional[str]    perplexity = None  # type: Optional[float]
    multiscale = None  # type: Optional[bool]
    exaggeration = None  # type: Optional[float]initialization_method = None    # type: Optional[str]
    initialization = None  # type: Optional[np.ndarray]
    affinities = None  # type: Optional[openTSNE.affinity.Affinities]
    tsne_embedding = None  # type: Optional[manifold.TSNEModel]
    iterations_done = 0  # type: int

    # These attributes need not be set by the widget
    tsne = None  # type: Optional[manifold.TSNE]# `effective_data` stores the current working matrix which should be used
    # for any steps depending on the data matrix. For instance, normalization
    # should use the effective data (presumably the original data), and set it
    # to the normalized version upon completion. This can then later be used for
    # PCA preprocessing
    effective_data = None           # type: Optional[Table]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Set `effective_data` to `data` if provided and effective data is not
        # provided
        if self.effective_data is None and self.data is not None:
            self.effective_data = self.data

    class ValidationError(ValueError):
        pass

    def validate(self) -> "Task":
        def error(msg):
            raise Task.ValidationError(msg)

        if self.data is None and self.distance_matrix is None:
            error("Both `distance_matrix` and `data` cannot be `None`")

        if self.distance_matrix is not None:
            if self.distance_metric != "precomputed":
                error(
                    "`distance_metric` must be set to `precomputed` when using "
                    "a distance matrix"
                )
            if self.initialization_method != "spectral":
                error(
                    "`initialization_method` must be set to `spectral` when "
                    "using a distance matrix"
                )

        if self.distance_matrix is None:
            if self.distance_metric == "precomputed":
                error(
                    "`distance_metric` cannot be set to `precomputed` when no "
                    "distance matrix is provided"
                )

        if self.data is not None and self.data.is_sparse():
            if self.normalize:
                error("Data normalization is not supported for sparse data")

        return self


def apply_tsne_preprocessing(tsne, data):
    return tsne.preprocess(data)


def data_normalization(data):
    normalization = preprocess.Normalize()
    return normalization(data)


def pca_preprocessing(data, n_components):
    projector = PCA(n_components=n_components, random_state=0)
    model = projector(data)
    return model(data)


def prepare_tsne_obj(n_samples: int, initialization_method: str,
                     distance_metric: str, perplexity: float,
                     multiscale: bool, exaggeration: float):
    """Automatically determine the best parameters for the given data set."""
    # Compute perplexity settings for multiscale
    if multiscale:
        perplexity = min((n_samples - 1) / 3, 50), min((n_samples - 1) / 3, 500)
    else:
        perplexity = perplexity

    return manifold.TSNE(
        n_components=2,
        initialization=initialization_method,
        metric=distance_metric,
        perplexity=perplexity,
        multiscale=multiscale,
        exaggeration=exaggeration,
        random_state=0,
    )


class TSNERunner:
    @staticmethod
    def compute_tsne_preprocessing(task: Task, state: TaskState, **_) -> None:
        state.set_status("Preprocessing data...")
        task.preprocessed_data = apply_tsne_preprocessing(task.tsne, task.effective_data)
        task.effective_data = task.preprocessed_data
        state.set_partial_result(("preprocessed_data", task))

    @staticmethod
    def compute_normalization(task: Task, state: TaskState, **_) -> None:
        state.set_status("Normalizing data...")
        task.normalized_data = data_normalization(task.effective_data)
        task.effective_data = task.normalized_data
        state.set_partial_result(("normalized_data", task))

    @staticmethod
    def compute_pca(task: Task, state: TaskState, **_) -> None:
        # Perform PCA preprocessing
        state.set_status(__("state.computing_PCA"))
        pca_projection = pca_preprocessing(task.effective_data, task.pca_components)
        # Apply t-SNE's preprocessors to the data
        task.pca_projection = task.tsne.preprocess(pca_projection)
        task.effective_data = task.pca_projection
        state.set_partial_result(("pca_projection", task))

    @staticmethod
    def compute_initialization(task: Task, state: TaskState, **_) -> None:
        # Prepare initial positions for t-SNE
        state.set_status(__("state.preparing_initialization"))
        if task.initialization_method == "pca":
            x = task.effective_data.X
        elif task.initialization_method == "spectral":
            assert task.affinities is not None
            x = task.affinities.P
        else:
            raise RuntimeError(
                f"Unrecognized initialization scheme `{task.initialization_method}`!"
            )
        task.initialization = task.tsne.compute_initialization(x)
        state.set_partial_result(("initialization", task))

    @staticmethod
    def compute_affinities(task: Task, state: TaskState, **_) -> None:
        state.set_status(__("state.find_nearest_neighbors"))

        if task.distance_metric == "precomputed":
            assert task.distance_matrix is not None
            x = task.distance_matrix
        else:
            assert task.data is not None
            assert task.effective_data is not None
            x = task.effective_data.X

        task.affinities = task.tsne.compute_affinities(x)
        state.set_partial_result(("affinities", task))

    @staticmethod
    def compute_tsne(task: Task, state: TaskState, progress_callback=None) -> None:
        tsne = task.tsne

        state.set_status(__("state.running_optimization"))

        # If this the first time we're computing t-SNE (otherwise we may just
        # be resuming optimization), we have to assemble the tsne object
        if task.tsne_embedding is None:
            # Assemble a t-SNE embedding object and convert it to a TSNEModel
            task.tsne_embedding = tsne.prepare_embedding(
                task.affinities, task.initialization
            )

            if task.distance_metric == "precomputed":
                x = task.distance_matrix
            else:
                assert task.effective_data is not None
                x = task.effective_data

            task.tsne_embedding = tsne.convert_embedding_to_model(
                x, task.tsne_embedding
            )
            state.set_partial_result(("tsne_embedding", task))

            if state.is_interruption_requested():
                return

        total_iterations_needed = tsne.early_exaggeration_iter + tsne.n_iter

        def run_optimization(tsne_params: dict, iterations_needed: int) -> bool:
            """Run t-SNE optimization phase. Return value indicates whether the
            optimization was interrupted."""
            while task.iterations_done < iterations_needed:
                # Step size can't be larger than the remaining number of iterations
                step_size = min(_STEP_SIZE, iterations_needed - task.iterations_done)
                task.tsne_embedding = task.tsne_embedding.optimize(
                    step_size, **tsne_params
                )
                task.iterations_done += step_size
                state.set_partial_result(("tsne_embedding", task))
                if progress_callback is not None:
                    # The current iterations must be divided by the total
                    # number of iterations, not the number of iterations in the
                    # current phase (iterations_needed)
                    progress_callback(task.iterations_done / total_iterations_needed)

                if state.is_interruption_requested():
                    return True

        # Run early exaggeration phase
        was_interrupted = run_optimization(
            dict(exaggeration=tsne.early_exaggeration, momentum=0.8, inplace=False),
            iterations_needed=tsne.early_exaggeration_iter,
        )
        if was_interrupted:
            return
        # Run regular optimization phase
        run_optimization(
            dict(exaggeration=tsne.exaggeration, momentum=0.8, inplace=False),
            iterations_needed=total_iterations_needed,
        )

    @classmethod
    def run(cls, task: Task, state: TaskState) -> Task:
        task.validate()

        # Assign weights to each job indicating how much time will be spent on each
        weights = {"preprocessing": 1, "normalization": 1, "pca": 1, "init": 1, "aff": 25, "tsne": 50}
        total_weight = sum(weights.values())

        # Prepare the tsne object and add it to the spec
        if task.distance_matrix is not None:
            n_samples = task.distance_matrix.shape[0]
        else:
            assert task.data is not None
            n_samples = task.data.X.shape[0]

        task.tsne = prepare_tsne_obj(
            n_samples,
            task.initialization_method,
            task.distance_metric,
            task.perplexity,
            task.multiscale,
            task.exaggeration,
        )

        job_queue = []
        # Add the tasks that still need to be run to the job queue
        if task.distance_metric != "precomputed":
            task.effective_data = task.data
            if task.preprocessed_data is None:
                job_queue.append((cls.compute_tsne_preprocessing, weights["preprocessing"]))

            if task.normalize and task.normalized_data is None:
                job_queue.append((cls.compute_normalization, weights["normalization"]))

            if task.use_pca_preprocessing and task.pca_projection is None:
                job_queue.append((cls.compute_pca, weights["pca"]))

        if task.affinities is None:
            job_queue.append((cls.compute_affinities, weights["aff"]))

        if task.initialization is None:
            job_queue.append((cls.compute_initialization, weights["init"]))

        total_iterations = task.tsne.early_exaggeration_iter + task.tsne.n_iter
        if task.tsne_embedding is None or task.iterations_done < total_iterations:
            job_queue.append((cls.compute_tsne, weights["tsne"]))

        job_queue = [(partial(f, task, state), w) for f, w in job_queue]

        # Ensure the effective data is set to the appropriate, potentially
        # precomputed matrix
        task.effective_data = task.data
        if task.preprocessed_data is not None:
            task.effective_data = task.preprocessed_data
        if task.normalize and task.normalized_data is not None:
            task.effective_data = task.normalized_data
        if task.use_pca_preprocessing and task.pca_projection is not None:
            task.effective_data = task.pca_projection

        # Figure out the total weight of the jobs
        job_weight = sum(j[1] for j in job_queue)
        progress_done = total_weight - job_weight
        for job, job_weight in job_queue:

            def _progress_callback(val):
                state.set_progress_value(
                    (progress_done + val * job_weight) / total_weight * 100
                )

            if state.is_interruption_requested():
                return task

            # Execute the job
            job(progress_callback=_progress_callback)
            # Update the progress bar according to the weights assigned to
            # each job
            progress_done += job_weight
            state.set_progress_value(progress_done / total_weight * 100)

        return task


class OWtSNEGraph(OWScatterPlotBase):
    def update_coordinates(self):
        super().update_coordinates()
        if self.scatterplot_item is not None:
            self.view_box.setAspectLocked(True, 1)


class invalidated:
    # pylint: disable=invalid-name
    preprocessed_data = normalized_data = pca_projection = initialization = \
        affinities = tsne_embedding = False

    def __set__(self, instance, value):
        # `self._invalidate = True` should invalidate everything
        self.preprocessed_data = value
        self.normalized_data = value
        self.pca_projection = value
        self.initialization = value
        self.affinities = value
        self.tsne_embedding = value

    def __bool__(self):
        # If any of the values are invalidated, this should return true
        return (
            self.preprocessed_data or self.normalized_data or self.pca_projection or
            self.initialization or self.affinities or self.tsne_embedding
        )

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, ", ".join(
            "=".join([k, str(getattr(self, k))])
            for k in ["preprocessed_data", "normalized_data", "pca_projection",
                      "initialization", "affinities", "tsne_embedding"]
        ))


class OWtSNE(OWDataProjectionWidget, ConcurrentWidgetMixin):
    name = __("name")
    description = __("desc")
    icon = "icons/TSNE.svg"
    priority = 920
    keywords = "t-sne, tsne"

    settings_version = 4
    perplexity = ContextSetting(30)
    multiscale = ContextSetting(False)
    exaggeration = ContextSetting(1)
    initialization_method_idx = ContextSetting(0)
    distance_metric_idx = ContextSetting(0)

    normalize = ContextSetting(True)
    use_pca_preprocessing = ContextSetting(True)
    pca_components = ContextSetting(_DEFAULT_PCA_COMPONENTS)

    GRAPH_CLASS = OWtSNEGraph
    graph = SettingProvider(OWtSNEGraph)
    embedding_variables_names = ("t-SNE-x", "t-SNE-y")

    # Use `invalidated` descriptor, so we don't break the usage of
    # `_invalidated` in `OWDataProjectionWidget`, but still allow finer control
    # over which parts of the embedding to invalidate
    _invalidated = invalidated()

    class Inputs(OWDataProjectionWidget.Inputs):
        distances = Input("Distances", DistMatrix)

    class Information(OWDataProjectionWidget.Information):
        modified = Msg(__("msg.modify_settings"))

    class Warning(OWDataProjectionWidget.Warning):
        consider_using_pca_preprocessing = Msg(
            "The input data contains a large number of features, which may slow"
            " down t-SNE computation. Consider enabling PCA preprocessing."
        )

    class Error(OWDataProjectionWidget.Error):
        not_enough_rows = Msg(__("msg.input_data_rows"))
        not_enough_cols = Msg(__("msg.input_data_attributes"))
        constant_data = Msg(__("msg.input_data_constant"))
        no_valid_data = Msg(__("msg.no_projection_error"))

        distance_matrix_not_symmetric = Msg("Distance matrix is not symmetric")
        distance_matrix_too_small = Msg("Input matrix must be at least 2x2")

        dimension_mismatch = Msg("Data and distance dimensions do not match")

    def __init__(self):
        OWDataProjectionWidget.__init__(self)
        ConcurrentWidgetMixin.__init__(self)

        # Distance matrix from `Distances` signal
        self.distance_matrix = None       # type: Optional[DistMatrix]
        # Data table from the `self.matrix.row_items` (if present)
        self.distance_matrix_data = None  # type: Optional[Table]
        # Data table from `Data` signal
        self.signal_data = None           # type: Optional[Table]

        # Intermediate results
        self.preprocessed_data = None     # type: Optional[Table]
        self.normalized_data = None       # type: Optional[Table]
        self.pca_projection = None  # type: Optional[Table]
        self.initialization = None  # type: Optional[np.ndarray]
        self.affinities = None  # type: Optional[openTSNE.affinity.Affinities]
        self.tsne_embedding = None  # type: Optional[manifold.TSNEModel]
        self.iterations_done = 0  # type: int

    @property
    def normalize_(self):
        should_normalize = self.normalize
        if self.distance_matrix is not None:
            should_normalize = False
        if self.data is not None:
            if self.data.is_sparse():
                should_normalize = False
        return should_normalize

    @property
    def use_pca_preprocessing_(self):
        should_use_pca_preprocessing = self.use_pca_preprocessing
        if self.distance_matrix is not None:
            should_use_pca_preprocessing = False
        return should_use_pca_preprocessing

    @property
    def effective_data(self):
        return self.data.transform(Domain(self.effective_variables))

    def _add_controls(self):
        self._add_controls_start_box()
        super()._add_controls()

    def _add_controls_start_box(self):
        self.preprocessing_box = gui.vBox(self.controlArea, box=__("preprocessing"))
        self.normalize_cbx = gui.checkBox(
            self.preprocessing_box, self, "normalize", __("normalize_data"),
            callback=self._normalize_data_changed, stateWhenDisabled=False,
        )
        self.pca_preprocessing_cbx = gui.checkBox(
            self.preprocessing_box, self, "use_pca_preprocessing", __("apply_pca"),
            callback=self._pca_preprocessing_changed, stateWhenDisabled=False,
        )
        self.pca_component_slider = gui.hSlider(
            self.preprocessing_box, self, "pca_components", label=__("pca_components"),
            minValue=2, maxValue=_MAX_PCA_COMPONENTS, step=1,
            callback=self._pca_slider_changed,
        )

        self.parameter_box = gui.vBox(self.controlArea, box=__("parameters"))
        form = QFormLayout(
            labelAlignment=Qt.AlignLeft,
            formAlignment=Qt.AlignLeft,
            fieldGrowthPolicy=QFormLayout.AllNonFixedFieldsGrow,
        )

        self.initialization_combo = gui.comboBox(
            self.controlArea, self, "initialization_method_idx",
            items=[m[0] for m in INITIALIZATIONS],
            callback=self._invalidate_initialization,
        )
        form.addRow(__("initialization"), self.initialization_combo)

        self.distance_metric_combo = gui.comboBox(
            self.controlArea, self, "distance_metric_idx",
            items=[m[0] for m in DISTANCE_METRICS],
            callback=self._invalidate_affinities,
        )
        form.addRow(__("distance_metric"), self.distance_metric_combo)

        self.perplexity_spin = gui.spin(
            self.controlArea, self, "perplexity", 1, 500, step=1,
            alignment=Qt.AlignRight, addToLayout=False,
            callback=self._invalidate_affinities,
        )
        form.addRow(__("row.perplexity"), self.perplexity_spin)
        form.addRow(gui.checkBox(
            self.controlArea, self, "multiscale", label=__("row.global_structure"),
            callback=self._multiscale_changed, addToLayout=False
        ))

        sbe = gui.hBox(self.controlArea, False, addToLayout=False)
        gui.hSlider(
            sbe, self, "exaggeration", minValue=1, maxValue=4, step=0.25,
            intOnly=False, labelFormat="%.2f",
            callback=self._invalidate_tsne_embedding,
        )
        form.addRow(__("row.exaggeration"), sbe)

        self.parameter_box.layout().addLayout(form)

        self.run_button = gui.button(
            self.parameter_box, self, __("btn.start"), callback=self._toggle_run
        )

    # GUI control callbacks
    def _normalize_data_changed(self):
        # We only care about the normalization checkbox if there is no distance
        # matrix provided and if the data are not sparse. This is not user-
        # settable anyway, but is triggered when we programmatically
        # enable/disable the checkbox in`enable_controls`
        if self.distance_matrix is None and not self.data.is_sparse():
            self._invalidate_normalized_data()

    def _pca_preprocessing_changed(self):
        # We only care about the PCA checkbox if there is no distance
        # matrix provided. This is not user-settable anyway, but is triggered
        # when we programmatically enable/disable the checkbox in
        # `enable_controls`
        if self.distance_matrix is None:
            self.controls.pca_components.box.setEnabled(self.use_pca_preprocessing)

            self._invalidate_pca_projection()

            should_warn_pca = False
            if self.data is not None and not self.use_pca_preprocessing:
                if len(self.data.domain.attributes) >= _MAX_PCA_COMPONENTS:
                    should_warn_pca = True
            self.Warning.consider_using_pca_preprocessing(shown=should_warn_pca)

    def _pca_slider_changed(self):
        # We only care about the PCA slider if there is no distance
        # matrix provided. This is not user-settable anyway, but is triggered
        # when we programmatically enable/disable the checkbox in
        # `enable_controls`
        if self.distance_matrix is None:
            self._invalidate_pca_projection()

    def _multiscale_changed(self):
        form = self.parameter_box.layout().itemAt(0)
        assert isinstance(form, QFormLayout)
        form.labelForField(self.perplexity_spin).setDisabled(self.multiscale)
        self.controls.perplexity.setDisabled(self.multiscale)

        self._invalidate_affinities()

    # Invalidation cascade
    def _invalidate_preprocessed_data(self):
        self._invalidated.preprocessed_data = True
        self._invalidate_normalized_data()

    def _invalidate_normalized_data(self):
        self._invalidated.normalized_data = True
        self._invalidate_pca_projection()

    def _invalidate_pca_projection(self):
        self._invalidated.pca_projection = True
        self._invalidate_affinities()

    def _invalidate_affinities(self):
        self._invalidated.affinities = True
        self._invalidate_tsne_embedding()

    def _invalidate_initialization(self):
        self._invalidated.initialization = True
        self._invalidate_tsne_embedding()

    def _invalidate_tsne_embedding(self):
        self._invalidated.tsne_embedding = True
        self._stop_running_task()
        self._set_modified(True)

    def _stop_running_task(self):
        self.cancel()
        self.run_button.setText(__("btn.start"))

    def _set_modified(self, state):
        """Mark the widget (GUI) as containing modified state."""
        if self.data is None:
            # Does not apply when we have no data
            state = False
        self.Information.modified(shown=state)

    def check_data(self):
        self.Error.dimension_mismatch.clear()
        self.Error.not_enough_rows.clear()
        self.Error.not_enough_cols.clear()
        self.Error.no_valid_data.clear()
        self.Error.constant_data.clear()

        if self.data is None:
            return

        def error(err):
            err()
            self.data = None

        if (
            self.data is not None and self.distance_matrix is not None and
            len(self.data) != len(self.distance_matrix)
        ):
            error(self.Error.dimension_mismatch)

        # The errors below are relevant only if the distance matrix is not
        # provided
        if self.distance_matrix is not None:
            return

        if len(self.data) < 2:
            error(self.Error.not_enough_rows)

        elif len(self.data.domain.attributes) < 2:
            error(self.Error.not_enough_cols)

        elif not self.data.is_sparse():
            if np.all(~np.isfinite(self.data.X)):
                error(self.Error.no_valid_data)
            else:
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", "Degrees of freedom .*", RuntimeWarning)
                    if np.nan_to_num(np.nanstd(self.data.X, axis=0)).sum() == 0:
                        error(self.Error.constant_data)

    def check_distance_matrix(self):
        self.Error.distance_matrix_not_symmetric.clear()
        self.Error.distance_matrix_too_small.clear()

        if self.distance_matrix is None:
            return

        def error(err):
            err()
            self.distance_matrix = self.distance_matrix_data = None

        # Check for matrix validity
        if self.distance_matrix is not None:
            if not self.distance_matrix.is_symmetric():
                error(self.Error.distance_matrix_not_symmetric)
            elif len(self.distance_matrix) < 2:
                error(self.Error.distance_matrix_too_small)

    def get_embedding(self):
        if self.tsne_embedding is None:
            self.valid_data = None
            return None

        embedding = self.tsne_embedding.embedding.X
        self.valid_data = np.ones(len(embedding), dtype=bool)
        return embedding

    def _toggle_run(self):
        # If no data, there's nothing to do
        if self.data is None:
            return

        # Pause task
        if self.task is not None:
            self.cancel()
            self.run_button.setText(i18n.t("common.software.resume"))
            self.commit.deferred()
        # Resume task
        else:
            self.run()

    @Inputs.data
    def set_data(self, data):
        self.signal_data = data
        # Data checking will be performed in `handleNewSignals` since the data
        # can also be set from the `distance_matrix.row_items` if no additional
        # data is provided

    @Inputs.distances
    def set_distances(self, matrix: DistMatrix):
        had_distance_matrix = self.distance_matrix is not None
        prev_distance_matrix = self.distance_matrix

        self.distance_matrix = matrix
        self.distance_matrix_data = matrix.row_items if matrix is not None else None
        self.check_distance_matrix()

        # If there was no distance matrix before, but there is data now, invalidate
        if self.distance_matrix is not None and not had_distance_matrix:
            self._invalidated = True

        # If the new distance matrix is invalid or None, invalidate
        elif self.distance_matrix is None and had_distance_matrix:
            self._invalidated = True

        # If the distance matrix has changed, invalidate
        elif (
            had_distance_matrix and self.distance_matrix is not None and
            not array_equal(prev_distance_matrix, self.distance_matrix)
        ):
            self._invalidated = True

    def handleNewSignals(self):
        had_data = self.data is not None
        prev_data = self.effective_data if had_data else None

        self.cancel()  # clear any running jobs
        self.data = None
        self.closeContext()

        if self.signal_data is not None:
            self.data = self.signal_data
        elif self.distance_matrix_data is not None:
            self.data = self.distance_matrix_data

        self.check_data()

        # If we have any errors, there's something wrong with the inputs or
        # their combination, so we clear the graph and the outputs
        if len(self.Error.active) > 0:
            self.clear()
            self._invalidated = True
            # Set data to None so that the output signal will be cleared
            self.data = None
            self.init_attr_values()
            self.commit.now()
            return

        # We only invalidate based on data if there is no distance matrix, as
        # otherwise, the embedding will remain in-tact
        if self.distance_matrix is None:
            # If there was no data before, but there is data now, invalidate
            if self.data is not None and not had_data:
                self._invalidated = True

            # If the new data is invalid or None, invalidate
            elif self.data is None and had_data:
                self._invalidated = True

            # If the data table has changed, invalidate
            elif (
                had_data and self.data is not None and
                not array_equal(prev_data.X, self.effective_data.X)
            ):
                self._invalidated = True

        self.init_attr_values()
        self.openContext(self.data)
        self.enable_controls()

        if self._invalidated:
            self.clear()
            self.input_changed.emit(self.data)

        # We don't bother with the granular invalidation flags because
        # `super().handleNewSignals` will just set all of them to False or will
        # do nothing. However, it's important we remember its state because we
        # won't call `run` if needed. `run` also relies on the state of
        # `_invalidated` to properly set the intermediate values to None
        prev_invalidated = bool(self._invalidated)
        super().handleNewSignals()
        self._invalidated = prev_invalidated

        if self._invalidated:
            self.run()

    def init_attr_values(self):
        super().init_attr_values()

        if self.data is not None:
            n_attrs = len(self.data.domain.attributes)
            max_components = min(_MAX_PCA_COMPONENTS, n_attrs)
            should_use_pca = len(self.data.domain.attributes) > 10
        else:
            max_components = _MAX_PCA_COMPONENTS
            should_use_pca = False

        # We set this to the default number of components here, so it resets
        # properly, any previous settings will be restored from context
        # settings a little later
        self.controls.pca_components.setMaximum(max_components)
        self.controls.pca_components.setValue(_DEFAULT_PCA_COMPONENTS)

        self.exaggeration = 1
        self.normalize = True
        self.use_pca_preprocessing = should_use_pca
        self.distance_metric_idx = 0
        self.initialization_method_idx = 0

    def enable_controls(self):
        super().enable_controls()

        has_distance_matrix = self.distance_matrix is not None
        has_data = self.data is not None

        # When we disable controls in the form layout, we also want to ensure
        # the labels are disabled, to be consistent with the preprocessing box
        form = self.parameter_box.layout().itemAt(0)
        assert isinstance(form, QFormLayout)

        # Reset all tooltips and controls
        self.normalize_cbx.setDisabled(False)
        self.normalize_cbx.setToolTip("")

        self.pca_preprocessing_cbx.setDisabled(False)
        self.pca_preprocessing_cbx.setToolTip("")

        self.initialization_combo.setDisabled(False)
        self.initialization_combo.setToolTip("")
        form.labelForField(self.initialization_combo).setDisabled(False)

        self.distance_metric_combo.setDisabled(False)
        self.distance_metric_combo.setToolTip("")
        form.labelForField(self.distance_metric_combo).setDisabled(False)

        if has_distance_matrix:
            self.normalize_cbx.setDisabled(True)
            self.normalize_cbx.setToolTip(
                "Precomputed distances provided. Preprocessing is unnecessary!"
            )

            self.pca_preprocessing_cbx.setDisabled(True)
            self.pca_preprocessing_cbx.setToolTip(
                "Precomputed distances provided. Preprocessing is unnecessary!"
            )

            # Only spectral init is valid with a precomputed distance matrix
            spectral_init_idx = self.initialization_combo.findText("Spectral")
            self.initialization_combo.setCurrentIndex(spectral_init_idx)
            self.initialization_combo.setDisabled(True)
            self.initialization_combo.setToolTip(
                "Only spectral intialization is supported with precomputed "
                "distance matrices."
            )
            form.labelForField(self.initialization_combo).setDisabled(True)

            self.distance_metric_combo.setDisabled(True)
            self.distance_metric_combo.setCurrentIndex(-1)
            self.distance_metric_combo.setToolTip(
                "Precomputed distances provided."
            )
            form.labelForField(self.distance_metric_combo).setDisabled(True)

        # Normalization isn't supported on sparse data, as this would
        # require centering and normalizing the matrix
        if not has_distance_matrix and has_data and self.data.is_sparse():
            self.normalize_cbx.setDisabled(True)
            self.normalize_cbx.setToolTip(
                __("tooltip_not_support")
            )

        # Disable slider parent, because we want to disable the labels too
        self.pca_component_slider.parent().setEnabled(self.use_pca_preprocessing_)

        # Disable the perplexity spin box if multiscale is turned on
        self.perplexity_spin.setDisabled(self.multiscale)
        form.labelForField(self.perplexity_spin).setDisabled(self.multiscale)

    def run(self):
        # Reset invalidated values as indicated by the flags
        if self._invalidated.preprocessed_data:
            self.preprocessed_data = None
        if self._invalidated.normalized_data:
            self.normalized_data = None
        if self._invalidated.pca_projection:
            self.pca_projection = None
        if self._invalidated.affinities:
            self.affinities = None
        if self._invalidated.initialization:
            self.initialization = None
        if self._invalidated.tsne_embedding:
            self.iterations_done = 0
            self.tsne_embedding = None

        self._set_modified(False)
        self._invalidated = False

        # When the data is invalid, it is set to `None` and an error is set,
        # therefore it would be erroneous to clear the error here
        if self.data is not None:
            self.run_button.setText(__("btn.stop"))

        # Cancel current running task
        self.cancel()

        if self.data is None and self.distance_matrix is None:
            return

        initialization_method = INITIALIZATIONS[self.initialization_method_idx][1]
        distance_metric = DISTANCE_METRICS[self.distance_metric_idx][1]
        if self.distance_matrix is not None:
            distance_metric = "precomputed"
            initialization_method = "spectral"

        task = Task(
            data=self.data,
            distance_matrix=self.distance_matrix,
            # Preprocessed data
            preprocessed_data=self.preprocessed_data,
            # Normalization
            normalize=self.normalize_,
            normalized_data=self.normalized_data,
            # PCA preprocessing
            use_pca_preprocessing=self.use_pca_preprocessing_,
            pca_components=self.pca_components,
            pca_projection=self.pca_projection,
            # t-SNE parameters
            initialization_method=initialization_method,
            initialization=self.initialization,
            distance_metric=distance_metric,
            perplexity=self.perplexity,
            multiscale=self.multiscale,
            exaggeration=self.exaggeration,
            affinities=self.affinities,
            # Misc
            tsne_embedding=self.tsne_embedding,
            iterations_done=self.iterations_done,
        )
        return self.start(TSNERunner.run, task)

    def __ensure_task_same_for_preprocessing(self, task: Task):
        if task.distance_metric != "precomputed":
            assert task.data is self.data
            assert isinstance(task.preprocessed_data, Table) and \
                len(task.preprocessed_data) == len(self.data)

    def __ensure_task_same_for_normalization(self, task: Task):
        assert task.normalize == self.normalize_
        if task.normalize and task.distance_metric != "precomputed":
            assert task.data is self.data
            assert isinstance(task.normalized_data, Table) and \
                len(task.normalized_data) == len(self.data)

    def __ensure_task_same_for_pca(self, task: Task):
        assert task.use_pca_preprocessing == self.use_pca_preprocessing_
        if task.use_pca_preprocessing and task.distance_metric != "precomputed":
            assert task.data is self.data
        assert task.pca_components == self.pca_components
        assert isinstance(task.pca_projection, Table) and \
               len(task.pca_projection) == len(self.data)

    def __ensure_task_same_for_initialization(self, task: Task):
        if self.distance_matrix is not None:
            n_samples = self.distance_matrix.shape[0]
        else:
            initialization_method = INITIALIZATIONS[self.initialization_method_idx][1]
            # If distance matrix is provided, the control value will be set to
            # whatever it was from the context, but we will use `spectral`
            assert task.initialization_method == initialization_method
            assert self.data is not None
            n_samples = self.data.X.shape[0]
        assert isinstance(task.initialization, np.ndarray) and \
               len(task.initialization) == n_samples

    def __ensure_task_same_for_affinities(self, task: Task):
        assert task.perplexity == self.perplexity
        assert task.multiscale == self.multiscale
        distance_metric = DISTANCE_METRICS[self.distance_metric_idx][1]
        # Precomputed distances will never match the combo box value
        if task.distance_metric != "precomputed":
            assert task.distance_metric == distance_metric

    def __ensure_task_same_for_embedding(self, task: Task):
        assert task.exaggeration == self.exaggeration
        if self.distance_matrix is not None:
            n_samples = self.distance_matrix.shape[0]
        else:
            assert self.data is not None
            n_samples = self.data.X.shape[0]
        assert isinstance(task.tsne_embedding, manifold.TSNEModel) and \
               len(task.tsne_embedding.embedding) == n_samples

    def on_partial_result(self, value):
        # type: (Tuple[str, Task]) -> None
        which, task = value

        if which == "preprocessed_data":
            self.__ensure_task_same_for_preprocessing(task)
            self.preprocessed_data = task.preprocessed_data
        elif which == "normalized_data":
            self.__ensure_task_same_for_preprocessing(task)
            self.__ensure_task_same_for_normalization(task)
            self.normalized_data = task.normalized_data
        elif which == "pca_projection":
            self.__ensure_task_same_for_preprocessing(task)
            self.__ensure_task_same_for_normalization(task)
            self.__ensure_task_same_for_pca(task)
            self.pca_projection = task.pca_projection
        elif which == "initialization":
            self.__ensure_task_same_for_preprocessing(task)
            self.__ensure_task_same_for_normalization(task)
            self.__ensure_task_same_for_pca(task)
            self.__ensure_task_same_for_initialization(task)
            self.initialization = task.initialization
        elif which == "affinities":
            self.__ensure_task_same_for_preprocessing(task)
            self.__ensure_task_same_for_normalization(task)
            self.__ensure_task_same_for_pca(task)
            self.__ensure_task_same_for_affinities(task)
            self.affinities = task.affinities
        elif which == "tsne_embedding":
            self.__ensure_task_same_for_preprocessing(task)
            self.__ensure_task_same_for_normalization(task)
            self.__ensure_task_same_for_pca(task)
            self.__ensure_task_same_for_initialization(task)
            self.__ensure_task_same_for_affinities(task)
            self.__ensure_task_same_for_embedding(task)

            prev_embedding, self.tsne_embedding = self.tsne_embedding, task.tsne_embedding
            self.iterations_done = task.iterations_done
            # If this is the first partial result we've gotten, we've got to
            # set up the plot
            if prev_embedding is None:
                self.setup_plot()
            # Otherwise, just update the point positions
            else:
                self.graph.update_coordinates()
                self.graph.update_density()
        else:
            raise RuntimeError(
                "Unrecognized partial result called with `%s`" % which
            )

    def on_done(self, task):
        # type: (Task) -> None
        self.run_button.setText(__("btn.start"))
        # NOTE: All of these have already been set by on_partial_result,
        # we double-check that they are aliases
        if task.preprocessed_data is not None:
            self.__ensure_task_same_for_preprocessing(task)
            assert task.preprocessed_data is self.preprocessed_data
        if task.normalized_data is not None:
            self.__ensure_task_same_for_normalization(task)
            assert task.normalized_data is self.normalized_data
        if task.pca_projection is not None:
            self.__ensure_task_same_for_pca(task)
            assert task.pca_projection is self.pca_projection
        if task.initialization is not None:
            self.__ensure_task_same_for_initialization(task)
            assert task.initialization is self.initialization
        if task.affinities is not None:
            assert task.affinities is self.affinities
        if task.tsne_embedding is not None:
            self.__ensure_task_same_for_embedding(task)
            assert task.tsne_embedding is self.tsne_embedding

        self.commit.deferred()

    def cancel(self):
        self.run_button.setText(__("btn.start"))
        super().cancel()

    def clear(self):
        """Clear widget state. Note that this doesn't clear the data."""
        super().clear()
        self.cancel()
        self.preprocessed_data = None
        self.normalized_data = None
        self.pca_projection = None
        self.initialization = None
        self.affinities = None
        self.tsne_embedding = None
        self.iterations_done = 0

    def onDeleteWidget(self):
        self.clear()
        self.data = None
        self.shutdown()
        super().onDeleteWidget()

    @classmethod
    def migrate_settings(cls, settings, version):
        if version < 3:
            if "selection_indices" in settings:
                settings["selection"] = settings["selection_indices"]
        if version < 4:
            settings.pop("max_iter", None)

    @classmethod
    def migrate_context(cls, context, version):
        if version < 3:
            values = context.values
            values["attr_color"] = values["graph"]["attr_color"]
            values["attr_size"] = values["graph"]["attr_size"]
            values["attr_shape"] = values["graph"]["attr_shape"]
            values["attr_label"] = values["graph"]["attr_label"]


if __name__ == "__main__":
    import sys

    data = Table(sys.argv[1] if len(sys.argv) > 1 else "iris")
    from Orange.distance import Euclidean
    dist_matrix = Euclidean(data, normalize=True)
    WidgetPreview(OWtSNE).run(
        set_distances=dist_matrix,
        set_subset_data=data[np.random.choice(len(data), 10)],
    )



import warnings

import sklearn.ensemble as skl_ensemble

from Orange.base import SklLearner
from Orange.classification.base_classification import (
    SklLearnerClassification, SklModelClassification
)
from Orange.regression.base_regression import (
    SklLearnerRegression, SklModelRegression
)
from Orange.util import OrangeDeprecationWarning


__all__ = ['SklAdaBoostClassificationLearner', 'SklAdaBoostRegressionLearner']


class SklAdaBoostClassifier(SklModelClassification):
    pass


def base_estimator_deprecation():
    warnings.warn(
        "`base_estimator` is deprecated (to be removed in 3.39): use `estimator` instead.",
        OrangeDeprecationWarning, stacklevel=3)


class SklAdaBoostClassificationLearner(SklLearnerClassification):
    __wraps__ = skl_ensemble.AdaBoostClassifier
    __returns__ = SklAdaBoostClassifier
    supports_weights = True

    def __init__(self, estimator=None, n_estimators=50, learning_rate=1.,
                 algorithm='SAMME.R', random_state=None, preprocessors=None,
                 base_estimator="deprecated"):
        if base_estimator != "deprecated":
            base_estimator_deprecation()
            estimator = base_estimator
        del base_estimator
        from Orange.modelling import Fitter
        # If fitter, get the appropriate Learner instance
        if isinstance(estimator, Fitter):
            estimator = estimator.get_learner(
                estimator.CLASSIFICATION)
        # If sklearn learner, get the underlying sklearn representation
        if isinstance(estimator, SklLearner):
            estimator = estimator.__wraps__(**estimator.params)
        super().__init__(preprocessors=preprocessors)
        self.params = vars()


class SklAdaBoostRegressor(SklModelRegression):
    pass


class SklAdaBoostRegressionLearner(SklLearnerRegression):
    __wraps__ = skl_ensemble.AdaBoostRegressor
    __returns__ = SklAdaBoostRegressor
    supports_weights = True

    def __init__(self, estimator=None, n_estimators=50, learning_rate=1.,
                 loss='linear', random_state=None, preprocessors=None,
                 base_estimator="deprecated"):
        if base_estimator != "deprecated":
            base_estimator_deprecation()
            estimator = base_estimator
        del base_estimator
        from Orange.modelling import Fitter
        # If fitter, get the appropriate Learner instance
        if isinstance(estimator, Fitter):
            estimator = estimator.get_learner(
                estimator.REGRESSION)
        # If sklearn learner, get the underlying sklearn representation
        if isinstance(estimator, SklLearner):
            estimator = estimator.__wraps__(**estimator.params)
        super().__init__(preprocessors=preprocessors)
        self.params = vars()



import math
from typing import List, Callable
from xml.sax.saxutils import escape

import numpy as np
import scipy.stats as ss
from scipy.stats import linregress
from sklearn.neighbors import NearestNeighbors
from sklearn.metrics import r2_score

from AnyQt.QtCore import Qt, QTimer, QPointF
from AnyQt.QtGui import QColor, QFont
from AnyQt.QtWidgets import QGroupBox

import pyqtgraph as pg

from orangewidget.utils.combobox import ComboBoxSearch

from Orange.data import Table, Domain, DiscreteVariable, Variable
from Orange.data.sql.table import SqlTable, AUTO_DL_LIMIT
from Orange.preprocess.score import ReliefF, RReliefF

from Orange.widgets import gui, report
from Orange.widgets.io import MatplotlibFormat, MatplotlibPDFFormat
from Orange.widgets.settings import (
    Setting, ContextSetting, SettingProvider, IncompatibleContext)
from Orange.widgets.utils import get_variable_values_sorted
from Orange.widgets.utils.itemmodels import DomainModel
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.visualize.owscatterplotgraph import OWScatterPlotBase, \
    ScatterBaseParameterSetter
from Orange.widgets.visualize.utils.vizrank import VizRankDialogAttrPair, \
    VizRankMixin
from Orange.widgets.visualize.utils.customizableplot import Updater
from Orange.widgets.visualize.utils.widget import OWDataProjectionWidget
from Orange.widgets.widget import AttributeList, Msg, Input, Output
from Orange.i18n_config import *


def __(key):
    return i18n.t("widget.visualize.visualize.owscatterplot." + key)


class ScatterPlotVizRank(VizRankDialogAttrPair):
    minK = 10

    def compute_score(self, state):
        # pylint: disable=invalid-unary-operand-type
        attrs = [self.attr_order[i] for i in state]
        data = self.data
        data = data.transform(Domain(attrs, self.attr_color))
        data = data[~np.isnan(data.X).any(axis=1) & ~np.isnan(data.Y).T]
        if len(data) < self.minK:
            return None
        n_neighbors = min(self.minK, len(data) - 1)
        knn = NearestNeighbors(n_neighbors=n_neighbors).fit(data.X)
        ind = knn.kneighbors(return_distance=False)
        if data.domain.has_discrete_class:
            return -np.sum(data.Y[ind] == data.Y.reshape(-1, 1)) / \
                   n_neighbors / len(data.Y)
        else:
            return -r2_score(data.Y, np.mean(data.Y[ind], axis=1)) * \
                   (len(data.Y) / len(self.data))

    def score_attributes(self):
        assert self.attr_color is not None
        attrs = [
            v
            for v in self.attrs  # same attributes that are in xy combos
            if v is not self.attr_color and v.is_primitive()
        ]
        domain = Domain(attributes=attrs, class_vars=self.attr_color)
        data = self.data.transform(domain)
        relief = ReliefF if isinstance(domain.class_var, DiscreteVariable) \
            else RReliefF
        weights = relief(
            n_iterations=100, k_nearest=self.minK, random_state=0)(data)
        attrs = sorted(zip(weights, domain.attributes),
                       key=lambda x: (-x[0], x[1].name))
        return [a for _, a in attrs]


class ParameterSetter(ScatterBaseParameterSetter):
    DEFAULT_LINE_WIDTH = 3
    DEFAULT_LINE_ALPHA = 255

    def __init__(self, master):
        super().__init__(master)
        self.reg_line_label_font = QFont()
        self.reg_line_settings = {
            Updater.WIDTH_LABEL: self.DEFAULT_LINE_WIDTH,
            Updater.ALPHA_LABEL: self.DEFAULT_LINE_ALPHA,
            Updater.STYLE_LABEL: Updater.DEFAULT_LINE_STYLE,
        }

    def update_setters(self):
        super().update_setters()
        self.initial_settings[self.LABELS_BOX].update({
            self.AXIS_TITLE_LABEL: self.FONT_SETTING,
            self.AXIS_TICKS_LABEL: self.FONT_SETTING,
            self.LINE_LAB_LABEL: self.FONT_SETTING
        })
        self.initial_settings[self.PLOT_BOX] = {}
        self.initial_settings[self.PLOT_BOX][self.LINE_LABEL] = {
            Updater.WIDTH_LABEL: (range(1, 10), self.DEFAULT_LINE_WIDTH),
            Updater.ALPHA_LABEL: (range(0, 255, 5), self.DEFAULT_LINE_ALPHA),
            Updater.STYLE_LABEL: (list(Updater.LINE_STYLES),
                                  Updater.DEFAULT_LINE_STYLE),
        }

        def update_lines(**settings):
            self.reg_line_settings.update(**settings)
            Updater.update_inf_lines(self.reg_line_items,
                                     **self.reg_line_settings)
            Updater.update_lines(self.ellipse_items,
                                 **self.reg_line_settings)
            self.master.update_reg_line_label_colors()

        def update_line_label(**settings):
            self.reg_line_label_font = \
                Updater.change_font(self.reg_line_label_font, settings)
            Updater.update_label_font(self.reg_line_label_items,
                                      self.reg_line_label_font)

        self._setters[self.LABELS_BOX][self.LINE_LAB_LABEL] = update_line_label
        self._setters[self.PLOT_BOX] = {self.LINE_LABEL: update_lines}

    @property
    def axis_items(self):
        return [value["item"] for value in
                self.master.plot_widget.plotItem.axes.values()]

    @property
    def reg_line_items(self):
        return self.master.reg_line_items

    @property
    def reg_line_label_items(self):
        return [line.label for line in self.master.reg_line_items
                if hasattr(line, "label")]

    @property
    def ellipse_items(self):
        return self.master.ellipse_items


class OWScatterPlotGraph(OWScatterPlotBase):
    show_reg_line = Setting(False)
    orthonormal_regression = Setting(False)
    show_ellipse = Setting(False)
    jitter_continuous = Setting(False)

    def __init__(self, scatter_widget, parent):
        super().__init__(scatter_widget, parent)
        self.parameter_setter = ParameterSetter(self)
        self.reg_line_items = []
        self.ellipse_items: List[pg.PlotCurveItem] = []

    def clear(self):
        super().clear()
        self.reg_line_items.clear()
        self.ellipse_items.clear()

    def update_coordinates(self):
        super().update_coordinates()
        self.update_axes()
        # Don't update_regression line here: update_coordinates is always
        # followed by update_point_props, which calls update_colors

    def update_colors(self):
        super().update_colors()
        self.update_regression_line()
        self.update_ellipse()

    def jitter_coordinates(self, x, y):
        def get_span(attr):
            if attr.is_discrete:
                # Assuming the maximal jitter size is 10, a span of 4 will
                # jitter by 4 * 10 / 100 = 0.4, so there will be no overlap
                return 4
            elif self.jitter_continuous:
                return None  # Let _jitter_data determine the span
            else:
                return 0  # No jittering
        span_x = get_span(self.master.attr_x)
        span_y = get_span(self.master.attr_y)
        if self.jitter_size == 0 or (span_x == 0 and span_y == 0):
            return x, y
        return self._jitter_data(x, y, span_x, span_y)

    def update_axes(self):
        for axis, var in self.master.get_axes().items():
            axis_item = self.plot_widget.plotItem.getAxis(axis)
            if var and var.is_discrete:
                ticks = [list(enumerate(get_variable_values_sorted(var)))]
                axis_item.setTicks(ticks)
            else:
                axis_item.setTicks(None)
            use_time = var and var.is_time
            self.plot_widget.plotItem.getAxis(axis).use_time(use_time)
            self.plot_widget.setLabel(axis=axis, text=var or "")
            if not var:
                self.plot_widget.hideAxis(axis)

    @staticmethod
    def _orthonormal_line(x, y, color, width, style=Qt.SolidLine):
        # https://en.wikipedia.org/wiki/Deming_regression, with δ=0.
        pen = pg.mkPen(color=color, width=width, style=style)
        xm = np.mean(x)
        ym = np.mean(y)
        sxx, sxy, _, syy = np.cov(x, y, ddof=1).flatten()

        if sxy != 0:  # also covers sxx != 0 and syy != 0
            slope = (syy - sxx + np.sqrt((syy - sxx) ** 2 + 4 * sxy ** 2)) \
                    / (2 * sxy)
            intercept = ym - slope * xm
            xmin = x.min()
            return pg.InfiniteLine(
                QPointF(xmin, xmin * slope + intercept),
                np.degrees(np.arctan(slope)),
                pen)
        elif (sxx == 0) == (syy == 0):  # both zero or non-zero -> can't draw
            return None
        elif sxx != 0:
            return pg.InfiniteLine(QPointF(x.min(), ym), 0, pen)
        else:
            return pg.InfiniteLine(QPointF(xm, y.min()), 90, pen)

    @staticmethod
    def _regression_line(x, y, color, width, style=Qt.SolidLine):
        min_x, max_x = np.min(x), np.max(x)
        if min_x == max_x:
            return None
        slope, intercept, rvalue, _, _ = linregress(x, y)
        angle = np.degrees(np.arctan(slope))
        start_y = min_x * slope + intercept
        l_opts = dict(color=color, position=0.85,
                      rotateAxis=(1, 0), movable=True)
        return pg.InfiniteLine(
            pos=QPointF(min_x, start_y), angle=angle,
            pen=pg.mkPen(color=color, width=width, style=style),
            label=f"r = {rvalue:.2f}", labelOpts=l_opts)

    def _add_line(self, x, y, color):
        width = self.parameter_setter.reg_line_settings[Updater.WIDTH_LABEL]
        alpha = self.parameter_setter.reg_line_settings[Updater.ALPHA_LABEL]
        style = self.parameter_setter.reg_line_settings[Updater.STYLE_LABEL]
        style = Updater.LINE_STYLES[style]
        color.setAlpha(alpha)
        if self.orthonormal_regression:
            line = self._orthonormal_line(x, y, color, width, style)
        else:
            line = self._regression_line(x, y, color, width, style)
        if line is None:
            return
        self.plot_widget.addItem(line)
        self.reg_line_items.append(line)

        if hasattr(line, "label"):
            Updater.update_label_font(
                [line.label], self.parameter_setter.reg_line_label_font
            )

    def update_reg_line_label_colors(self):
        for line in self.reg_line_items:
            if hasattr(line, "label"):
                color = 0.0 if self.class_density \
                    else line.pen.color().darker(175)
                line.label.setColor(color)

    def update_density(self):
        super().update_density()
        self.update_reg_line_label_colors()

    def update_regression_line(self):
        self._update_curve(self.reg_line_items,
                           self.show_reg_line,
                           self._add_line)
        self.update_reg_line_label_colors()

    def update_ellipse(self):
        self._update_curve(self.ellipse_items,
                           self.show_ellipse,
                           self._add_ellipse)

    def _update_curve(self, items: List, show: bool, add: Callable):
        for item in items:
            self.plot_widget.removeItem(item)
        items.clear()
        if not (show and self.master.can_draw_regression_line()):
            return
        x, y = self.master.get_coordinates_data()
        if x is None or len(x) < 2:
            return
        add(x, y, QColor("#505050"))
        if self.master.is_continuous_color() or self.palette is None \
                or len(self.palette) == 0:
            return
        c_data = self.master.get_color_data()
        if c_data is None:
            return
        c_data = c_data.astype(int)
        for val in range(c_data.max() + 1):
            mask = c_data == val
            if mask.sum() > 1:
                add(x[mask], y[mask], self.palette[val].darker(135))

    def _add_ellipse(self, x: np.ndarray, y: np.ndarray, color: QColor) -> np.ndarray:
        # https://github.com/ChristianGoueguel/HotellingEllipse/blob/master/R/ellipseCoord.R
        points = np.vstack([x, y]).T
        mu = np.mean(points, axis=0)
        cov = np.cov(*(points - mu).T)
        vals, vects = np.linalg.eig(cov)
        angle = math.atan2(vects[1, 0], vects[0, 0])
        matrix = np.array([[np.cos(angle), -np.sin(angle)],
                           [np.sin(angle), np.cos(angle)]])

        n = len(x)
        f = ss.f.ppf(0.95, 2, n - 2)
        f = f * 2 * (n - 1) / (n - 2)
        m = [np.pi * i / 100 for i in range(201)]
        cx = np.cos(m) * np.sqrt(vals[0] * f)
        cy = np.sin(m) * np.sqrt(vals[1] * f)

        pts = np.vstack([cx, cy])
        pts = matrix.dot(pts)
        cx = pts[0] + mu[0]
        cy = pts[1] + mu[1]

        width = self.parameter_setter.reg_line_settings[Updater.WIDTH_LABEL]
        alpha = self.parameter_setter.reg_line_settings[Updater.ALPHA_LABEL]
        style = self.parameter_setter.reg_line_settings[Updater.STYLE_LABEL]
        style = Updater.LINE_STYLES[style]
        color.setAlpha(alpha)

        pen = pg.mkPen(color=color, width=width, style=style)
        ellipse = pg.PlotCurveItem(cx, cy, pen=pen)
        self.plot_widget.addItem(ellipse)
        self.ellipse_items.append(ellipse)


class OWScatterPlot(OWDataProjectionWidget, VizRankMixin(ScatterPlotVizRank)):
    """Scatterplot visualization with explorative analysis and intelligent
    data visualization enhancements."""

    name = __('name')
    description = __("desc")
    icon = "icons/ScatterPlot.svg"
    priority = 140
    keywords = "scatter plot"

    class Inputs(OWDataProjectionWidget.Inputs):
        features = Input('Features', AttributeList, label=i18n.t("common.general.features"))

    class Outputs(OWDataProjectionWidget.Outputs):
        features = Output('Features', AttributeList, dynamic=False, label=i18n.t("common.general.features"))

    settings_version = 5
    auto_sample = Setting(True)
    attr_x = ContextSetting(None)
    attr_y = ContextSetting(None)
    tooltip_shows_all = Setting(True)

    GRAPH_CLASS = OWScatterPlotGraph
    graph = SettingProvider(OWScatterPlotGraph)
    embedding_variables_names = None

    class Warning(OWDataProjectionWidget.Warning):
        missing_coords = Msg(__("msg.plot_display_error"))

    class Information(OWDataProjectionWidget.Information):
        sampled_sql = Msg(__("msg.sql_table_showing"))
        missing_coords = Msg(__("msg.point_display_tip"))

    def __init__(self):
        self.attr_box: QGroupBox = None
        self.xy_model: DomainModel = None
        self.cb_attr_x: ComboBoxSearch = None
        self.cb_attr_y: ComboBoxSearch = None
        self.sampling: QGroupBox = None
        self._xy_invalidated: bool = True

        self.sql_data = None  # Orange.data.sql.table.SqlTable
        self.attribute_selection_list = None  # list of Orange.data.Variable
        self.__timer = QTimer(self, interval=1200)
        self.__timer.timeout.connect(self.add_data)
        super().__init__()

        # manually register Matplotlib file writers
        self.graph_writers = self.graph_writers.copy()
        for w in [MatplotlibFormat, MatplotlibPDFFormat]:
            self.graph_writers.append(w)

    def _add_controls(self):
        self._add_controls_axis()
        self._add_controls_sampling()
        super()._add_controls()
        self.gui.add_widget(self.gui.JitterNumericValues, self._effects_box)
        self.gui.add_widgets(
            [self.gui.ShowGridLines,
             self.gui.ToolTipShowsAll,
             self.gui.RegressionLine],
            self._plot_box)
        gui.checkBox(
            self._plot_box, self,
            value="graph.orthonormal_regression",
            label=__("checkbox.variable_independent"),
            callback=self.graph.update_regression_line,
            tooltip=__("checkbox.variable_independent_tip"),
            disabledBy=self.cb_reg_line)
        gui.checkBox(
            self._plot_box, self,
            value="graph.show_ellipse",
            label=__("show_ellipse"),
            tooltip="Hotelling's T² confidence ellipse (α=95%)",
            callback=self.graph.update_ellipse)

    def _add_controls_axis(self):
        common_options = dict(
            labelWidth=50, orientation=Qt.Horizontal, sendSelectedValue=True,
            contentsLength=12, searchable=True
        )
        self.attr_box = gui.vBox(self.controlArea, __("box.axes"),
                                 spacing=2 if gui.is_macstyle() else 8)
        dmod = DomainModel
        self.xy_model = DomainModel(dmod.MIXED, valid_types=dmod.PRIMITIVE)
        self.cb_attr_x = gui.comboBox(
            self.attr_box, self, "attr_x", label=__("label.axis_x"),
            callback=self.set_attr_from_combo,
            model=self.xy_model, **common_options,
        )
        self.cb_attr_y = gui.comboBox(
            self.attr_box, self, "attr_y", label=__("label.axis_y"),
            callback=self.set_attr_from_combo,
            model=self.xy_model, **common_options,
        )
        vizrank_box = gui.hBox(self.attr_box)
        button = self.vizrank_button(__("btn_informative_projection"))
        vizrank_box.layout().addWidget(button)
        self.vizrankSelectionChanged.connect(self.set_attr)

    def _add_controls_sampling(self):
        self.sampling = gui.auto_commit(
            self.controlArea, self, "auto_sample", "Sample", box=__("box.sample"),
            callback=self.switch_sampling, commit=lambda: self.add_data(1))
        self.sampling.setVisible(False)

    @property
    def effective_variables(self):
        return [self.attr_x, self.attr_y] if self.attr_x and self.attr_y else []

    @property
    def effective_data(self):
        eff_var = self.effective_variables
        if eff_var and self.attr_x.name == self.attr_y.name:
            eff_var = [self.attr_x]
        return self.data.transform(Domain(eff_var))

    def init_vizrank(self):
        err_msg = ""
        if self.data is None:
            err_msg = __("msg.input_empty")
        elif self.data.is_sparse():
            err_msg = __("msg.data_sparse")
        elif len(self.xy_model) < 3:
            err_msg = __("msg.rank_error")
        elif self.attr_color is None:
            err_msg = __("msg.select_color_variable")
        elif np.isnan(self.data.get_column(self.attr_color)).all():
            err_msg = __("msg.no_value_color_variable")
        if not err_msg:
            super().init_vizrank(self.data, list(self.xy_model), self.attr_color)
        else:
            self.disable_vizrank(err_msg)

    @OWDataProjectionWidget.Inputs.data
    def set_data(self, data):
        super().set_data(data)
        self.init_vizrank()

        def findvar(name, iterable):
            """Find a Orange.data.Variable in `iterable` by name"""
            for el in iterable:
                if isinstance(el, Variable) and el.name == name:
                    return el
            return None

        # handle restored settings from  < 3.3.9 when attr_* were stored
        # by name
        if isinstance(self.attr_x, str):
            self.attr_x = findvar(self.attr_x, self.xy_model)
        if isinstance(self.attr_y, str):
            self.attr_y = findvar(self.attr_y, self.xy_model)
        if isinstance(self.attr_label, str):
            self.attr_label = findvar(self.attr_label, self.gui.label_model)
        if isinstance(self.attr_color, str):
            self.attr_color = findvar(self.attr_color, self.gui.color_model)
        if isinstance(self.attr_shape, str):
            self.attr_shape = findvar(self.attr_shape, self.gui.shape_model)
        if isinstance(self.attr_size, str):
            self.attr_size = findvar(self.attr_size, self.gui.size_model)

    def check_data(self):
        super().check_data()
        self.__timer.stop()
        self.sampling.setVisible(False)
        self.sql_data = None
        if isinstance(self.data, SqlTable):
            if self.data.approx_len() < 4000:
                self.data = Table(self.data)
            else:
                self.Information.sampled_sql()
                self.sql_data = self.data
                data_sample = self.data.sample_time(0.8, no_cache=True)
                data_sample.download_data(2000, partial=True)
                self.data = Table(data_sample)
                self.sampling.setVisible(True)
                if self.auto_sample:
                    self.__timer.start()

        if self.data is not None and (len(self.data) == 0 or
                                      len(self.data.domain.variables) == 0):
            self.data = None

    def get_embedding(self):
        self.valid_data = None
        if self.data is None:
            return None

        x_data = self.get_column(self.attr_x, filter_valid=False)
        y_data = self.get_column(self.attr_y, filter_valid=False)
        if x_data is None or y_data is None:
            return None

        self.Warning.missing_coords.clear()
        self.Information.missing_coords.clear()
        self.valid_data = np.isfinite(x_data) & np.isfinite(y_data)
        if self.valid_data is not None and not np.all(self.valid_data):
            msg = self.Information if np.any(self.valid_data) else self.Warning
            msg.missing_coords(self.attr_x.name, self.attr_y.name)
        return np.vstack((x_data, y_data)).T

    # Tooltip
    def _point_tooltip(self, point_id, skip_attrs=()):
        point_data = self.data[point_id]
        xy_attrs = (self.attr_x, self.attr_y)
        text = "<br/>".join(
            escape('{} = {}'.format(var.name, point_data[var]))
            for var in xy_attrs)
        if self.tooltip_shows_all:
            others = super()._point_tooltip(point_id, skip_attrs=xy_attrs)
            if others:
                text = "<b>{}</b><br/><br/>{}".format(text, others)
        return text

    def can_draw_regression_line(self):
        return self.data is not None and \
               self.data.domain is not None and \
               self.attr_x is not None and self.attr_y is not None and \
               self.attr_x.is_continuous and \
               self.attr_y.is_continuous

    def add_data(self, time=0.4):
        if self.data and len(self.data) > 2000:
            self.__timer.stop()
            return
        data_sample = self.sql_data.sample_time(time, no_cache=True)
        if data_sample:
            data_sample.download_data(2000, partial=True)
            data = Table(data_sample)
            self.data = Table.concatenate((self.data, data), axis=0)
            self.handleNewSignals()

    def init_attr_values(self):
        super().init_attr_values()
        data = self.data
        domain = data.domain if data and len(data) else None
        self.xy_model.set_domain(domain)
        self.attr_x = self.xy_model[0] if self.xy_model else None
        self.attr_y = self.xy_model[1] if len(self.xy_model) >= 2 \
            else self.attr_x

    def switch_sampling(self):
        self.__timer.stop()
        if self.auto_sample and self.sql_data:
            self.add_data()
            self.__timer.start()

    @OWDataProjectionWidget.Inputs.data_subset
    def set_subset_data(self, subset_data):
        self.warning()
        if isinstance(subset_data, SqlTable):
            if subset_data.approx_len() < AUTO_DL_LIMIT:
                subset_data = Table(subset_data)
            else:
                self.warning(__("msg.data_subset_error"))
                subset_data = None
        super().set_subset_data(subset_data)

    # called when all signals are received, so the graph is updated only once
    def handleNewSignals(self):
        self.attr_box.setEnabled(True)
        if self.attribute_selection_list and self.data is not None and \
                self.data.domain is not None:
            self.attr_box.setEnabled(False)
            if all(attr in self.xy_model for attr in self.attribute_selection_list):
                self.attr_x, self.attr_y = self.attribute_selection_list
            else:
                self.attr_x, self.attr_y = None, None
        self._invalidated = self._invalidated or self._xy_invalidated
        self._xy_invalidated = False
        super().handleNewSignals()
        if self._domain_invalidated:
            self.graph.update_axes()
            self._domain_invalidated = False
        can_plot = self.can_draw_regression_line()
        self.cb_reg_line.setEnabled(can_plot)
        self.graph.controls.show_ellipse.setEnabled(can_plot)

    @Inputs.features
    def set_shown_attributes(self, attributes):
        if attributes and len(attributes) >= 2:
            self.attribute_selection_list = attributes[:2]
            self._xy_invalidated = self._xy_invalidated \
                or self.attr_x != attributes[0] \
                or self.attr_y != attributes[1]
        else:
            if self.attr_x is None or self.attr_y is None:
                # scenario happens when features input removed and features
                # were invalid or hidden and those attr_x and attr_h were None
                self.init_attr_values()
            self.attribute_selection_list = None

    def set_attr(self, attrs):
        if attrs != [self.attr_x,self.attr_y]:
            self.attr_x, self.attr_y = attrs
            self.attr_changed()

    def set_attr_from_combo(self):
        self.attr_changed()
        self.vizrankAutoSelect.emit([self.attr_x, self.attr_y])

    def attr_changed(self):
        can_plot = self.can_draw_regression_line()
        self.cb_reg_line.setEnabled(can_plot)
        self.graph.controls.show_ellipse.setEnabled(can_plot)
        self.setup_plot()
        self.commit.deferred()

    def get_axes(self):
        return {"bottom": self.attr_x, "left": self.attr_y}

    def colors_changed(self):
        super().colors_changed()
        self.init_vizrank()

    @gui.deferred
    def commit(self):
        super().commit()
        self.send_features()

    def send_features(self):
        features = [attr for attr in [self.attr_x, self.attr_y] if attr]
        self.Outputs.features.send(AttributeList(features) or None)

    def get_widget_name_extension(self):
        if self.data is not None:
            return "{} vs {}".format(self.attr_x.name, self.attr_y.name)
        return None

    def _get_send_report_caption(self):
        return report.render_items_vert((
            ("Color", self._get_caption_var_name(self.attr_color)),
            ("Label", self._get_caption_var_name(self.attr_label)),
            ("Shape", self._get_caption_var_name(self.attr_shape)),
            ("Size", self._get_caption_var_name(self.attr_size)),
            ("Jittering", (self.attr_x.is_discrete or
                           self.attr_y.is_discrete or
                           self.graph.jitter_continuous) and
             self.graph.jitter_size)))

    @classmethod
    def migrate_settings(cls, settings, version):
        if version < 2 and "selection" in settings and settings["selection"]:
            settings["selection_group"] = [(a, 1) for a in settings["selection"]]
        if version < 3:
            if "auto_send_selection" in settings:
                settings["auto_commit"] = settings["auto_send_selection"]
            if "selection_group" in settings:
                settings["selection"] = settings["selection_group"]
        if version < 5:
            if "graph" in settings and \
                    "jitter_continuous" not in settings["graph"]:
                settings["graph"]["jitter_continuous"] = True

    @classmethod
    def migrate_context(cls, context, version):
        values = context.values
        if version < 3:
            values["attr_color"] = values["graph"]["attr_color"]
            values["attr_size"] = values["graph"]["attr_size"]
            values["attr_shape"] = values["graph"]["attr_shape"]
            values["attr_label"] = values["graph"]["attr_label"]
        if version < 4:
            if values["attr_x"][1] % 100 == 1 or values["attr_y"][1] % 100 == 1:
                raise IncompatibleContext()


if __name__ == "__main__":  # pragma: no cover
    table = Table("iris")
    WidgetPreview(OWScatterPlot).run(set_data=table,
                                     set_subset_data=table[:30])
